import requests
import datetime

from ..utils import exit_util


def auth_user_password(user, password=None):
    try:
        expiration = datetime.datetime.now()
        if password:
            data = {'grant_type': user.grand_type,
                    'client_id': user.client_id,
                    'secret_id': user.secret,
                    'username': user.username,
                    'password': password}
            resp = requests.post('https://stepik.org/oauth2/token/', data)
        else:
            auth = requests.auth.HTTPBasicAuth(user.client_id, user.secret)
            data = {'grant_type': user.grand_type}
            resp = requests.post('https://stepik.org/oauth2/token/', data, auth=auth)

        assert resp.status_code < 300

        resp = resp.json()
        user.access_token = resp['access_token']
        user.expiration = expiration + datetime.timedelta(seconds=int(resp['expires_in']))
        user.refresh_token = None
        if 'refresh_token' in resp:
            user.refresh_token = resp['refresh_token']
        user.save()
    except AssertionError:
        exit_util("Check your authentication.")


def get_headers(user):
    return {'Authorization': 'Bearer ' + user.access_token, "content-type": "application/json"}


def refresh_client(user):
    try:
        data = {'grant_type': 'refresh_token',
                'client_id': user.client_id,
                'secret_id': user.secret,
                'refresh_token': user.refresh_token}

        resp = requests.post('https://stepik.org/oauth2/token/', data)

        assert resp.status_code < 300

        user.access_token = (resp.json())['access_token']
        user.refresh_token = (resp.json())['refresh_token']
        user.save()
    except AssertionError:
        return False

    return True



#!/usr/bin/env python
import click
import html2text
from pathlib import Path

from . import attempt_cache
from .course_cache import CourseCache
from .client import stepikclient
from .client.auth import auth_user_password
from .client.consts import STEPIK_HOST, GRAND_TYPE_CREDENTIALS
from .filemanager import FileManager
from .models.course import Course
from .models.lesson import Lesson
from .models.section import Section
from .models.user import User
from .navigation import prev_step, next_step, create_course_cache
from .settings import APP_FOLDER, COURSE_CACHE_FILE, CLIENT_ID, CLIENT_SECRET
from .utils import exit_util


@click.group()
@click.version_option()
def main():
    """
    The (unofficial) Stepik CLI for students\n
    A command line tool for submitting solutions to stepik.org
    """
    file_manager = FileManager()
    try:
        file_manager.create_dir(APP_FOLDER)
    except OSError:
        exit_util("Can't do anything. Insufficient permissions to folders.")


@main.command()
def auth():
    """
    Authenticate using your OAuth2 credentials.
    """
    click.echo("Enter your registration info from https://stepik.org/oauth2/applications/")

    try:
        user = User()
        user.grand_type = GRAND_TYPE_CREDENTIALS

        if not (CLIENT_ID and CLIENT_SECRET):
            user.client_id = click.prompt(text="Enter your client ID")
            user.secret = click.prompt(text="Enter your client secret")
            click.secho("Authenticating...")
        auth_user_password(user)
        user.save()

    except Exception as e:
        exit_util("\nError! Double-check that your client ID and client secret are correct." + str(e))

    click.secho("Authentication was successfull!", fg="green", bold=True, err=True)


@main.command("step")
@click.argument("link")
def step_cmd(link=None):
    """
    Navigate the current position to the step at the provided URL.
    """
    if link is not None:
        user = User()
        stepikclient.set_step(user, link)

@main.command()
@click.argument("dataset-path", type=Path)
@click.option("--step_id", help="step id")
@click.option("--attempt_id", help="attempt id")
def dataset(dataset_path, step_id=None, attempt_id=None):
    """
    Attempt a dataset challenge. Writes the downloaded dataset to the provided path.
    """
    user = User()
    try:
        attempt = stepikclient.download_dataset(user, dataset_path, step_id, attempt_id)
    except:
        exit_util("Unable to download dataset.")
    click.secho(str(attempt), bold=True, fg='green', err=True)


@main.command()
@click.argument("solution")
@click.option("-l", help="programming language")
@click.option("--step-id", help="step id")
@click.option("--attempt-id", help="attempt id")
def submit(solution=None, l=None, step_id=None, attempt_id=None):
    """
    Submit a solution to stepik. Use the contents of the provided file path.\n
    Specify the programming language via the -l option if your submission is code.\n
    If you are NOT submitting the solution to a dataset challenge, specify "text" to -l
    """
    if solution is not None:
        user = User()
        try:
            stepikclient.submit_code(user, solution, l, step_id, attempt_id)
        except:
            exit_util("Unable to submit solution. There was some sort of error.")


@main.command()
def lang():
    """
    Lists the available programming languages for the current step.\n
    Assumes the current step has a coding challenge.
    """
    user = User()
    current_step_id = attempt_cache.get_step_id()
    languages = stepikclient.get_languages_list(user, current_step_id)
    languages.sort()
    click.secho(", ".join(languages), bold=True, nl=False)


@main.command("next")
def next_cmd():
    """
    Navigate to the next step in a course.\n
    For the best navigation experience, you should set the course using the "course" command before using this command.\n
    Steps will be filtered according to the current step type. You can use the "type" command to change this.
    """
    user = User()
    if next_step(user, user.step_type):
        current_lesson = attempt_cache.get_lesson_id()
        current_pos = attempt_cache.get_current_position()
        message = "Switched to lesson {}, step {}".format(current_lesson, current_pos)
        color = "green"
    else:
        message = "Unable to switch to next step in this lesson."
        message += "\nIf you would like to proceed to the next lesson, check that you've set the course using the 'course' command."
        color = "red"

    click.secho(message, bold=True, fg=color, err=True)


@main.command()
def prev():
    """
    Navigate to the previous step in a course.\n
    For the best navigation experience, you should set the course using the "course" command before using this command.\n
    Steps will be filtered according to the current step type. You can use the "type" command to change this.
    """
    user = User()
    if prev_step(user, user.step_type):
        current_lesson = attempt_cache.get_lesson_id()
        current_pos = attempt_cache.get_current_position()
        message = "Switched to lesson {}, step {}".format(current_lesson, current_pos)
        color = "green"
    else:
        message = "Unable to switch to previous step in this lesson."
        message += "\nIf you would like to proceed to the previous lesson, check that you've set the course using the 'course' command."
        color = "red"

    click.secho(message, bold=True, fg=color, err=True)


@main.command("type")
@click.argument("step_type", type=click.Choice(['all', 'code', 'text', 'dataset'], case_sensitive=False), default='all')
def type_cmd(step_type="dataset"):
    """
    Filter for steps with this step type.\n
    When navigating using the "next" or "prev" commands, any steps that are not of this type will be skipped.
    """
    user = User()
    user.step_type = step_type
    user.save()

    if step_type == 'all':
        click.secho("Steps will not be filtered for their type when navigating.", bold=True, fg="green", err=True)
    else:
        click.secho("Steps that are not of type {} will be ignored when navigating.".format(step_type), bold=True, fg="green", err=True)


@main.command()
def current():
    """
    Display the URL and step ID of the current step.
    """
    data = attempt_cache.get_data()
    lesson_id = attempt_cache.get_lesson_id(data)
    if lesson_id is None:
        exit_util("You should first set the current step using the 'step' command.")

    step_position = attempt_cache.get_current_position(data)
    step_id = attempt_cache.get_step_id(data)

    click.secho(STEPIK_HOST + "lesson/{}/step/{}\t{}".format(lesson_id, step_position, step_id), bold=True, fg="green")


@main.command()
def text():
    """
    Display the contents of the current step.
    """
    user = User()

    step_id = attempt_cache.get_step_id()
    if step_id is None:
        exit_util("You should first set the current step using the 'step' command.")

    step = stepikclient.get_step(user, step_id)

    html = step['steps'][0]['block']['text']
    click.secho(html2text.html2text(html))


@main.command()
def courses():
    """
    Display a list of your enrolled courses and their course IDs.
    """
    courses_set = "\n".join(map(str, Course.all()))
    click.secho(courses_set)


def validate_id(ctx, param, value):
    if value < 1:
        raise click.BadParameter('Should be a positive integer greater than zero.')
    return value


@main.command('course')
@click.argument("course_id", type=click.INT, callback=validate_id)
@click.option(
    '--recache', is_flag=True, help=(
        "Force a recache of this course. Provide this flag if sections or lessons in "
        "the course have changed since your last cache of the course. Otherwise, "
        "you will have unexpected results when navigating via the next/prev commands."
    )
)
def course_cmd(course_id, recache=False):
    """
    Switch to the course that has the provided course ID.\n
    Cache the course for navigation purposes and display a description of the course.
    """
    user = User()
    course = Course.get(user, course_id)
    click.secho(str(course), bold=True)

    cache = create_course_cache(course)
    if cache.load(user) and not recache:
        click.secho("Retrieved course from cache.", fg='green', err=True)
    else:
        click.secho((
            "Caching course lessons...\nSince this is the first time you are navigating"
            " this course, this may take a minute."
        ), fg='red', bold=True, err=True)
        try:
            cache.update()
        except:
            exit_util("Unable to cache course. Do you have permission to view it?")
        if not cache.load(user):
            exit_util("Something went wrong. We were unable to cache this course.")

    click.secho(html2text.html2text(course.description))


_ENTITIES = {'course': Course, 'section': Section, 'lesson': Lesson}


def validate_entity(ctx, param, value):
    value = str(value).lower()
    if value not in _ENTITIES:
        raise click.BadParameter('Should be one from "course", "section", "lesson"')
    return value


@main.command('content')
@click.argument("entity", type=click.STRING, callback=validate_entity)
@click.argument("entity_id", type=click.INT, callback=validate_id)
def content_cmd(entity, entity_id):
    """
    View the content of a course, section, or lesson by its ID.\n
    Format:\n
        content course <course_id>\n
        content section <section_id>\n
        content lesson <lesson_id>\n
    """
    if entity not in _ENTITIES:
        return

    user = User()

    entity_class = _ENTITIES[entity]

    entity = entity_class.get(user, entity_id)

    click.secho(str(entity), bold=True)
    items = "\n".join(map(str, entity.items()))
    click.secho(items)



import re
import sys

import click


def exit_util(message, exit_code=1):
    click.secho(message, fg="red", bold=True, err=True)
    sys.exit(exit_code)


def get_lesson_id(step_url):
    match = re.search(r'lesson/.*?(\d+)/', step_url)
    if match is None:
        return match
    return match.group(1)


def get_step_id(step_url):
    match = re.search(r'step/(\d+)', step_url)
    if match is None:
        return 0
    return int(match.group(1))


def prepare_ids(ids):
    return "&".join(map(lambda id: "ids[]=" + str(id), ids))


def entities_loader(getter, user, key, ids, entity_class):
    has_next = True
    page_index = 1
    entities = list()

    while has_next:
        page = getter(user, ids, page_index)
        has_next = page['meta']['has_next']
        page_index += 1
        page = page[key]
        entities_list = list(map(lambda entity: entity_class(user, entity), page))
        entities.extend(entities_list)

    return entities


def all_entities_loader(getter, user, key, entity_class, **kwargs):
    has_next = True
    page_index = 1
    entities = list()

    while has_next:
        kwargs['page'] = page_index
        page = getter(user, **kwargs)
        has_next = page['meta']['has_next']
        page_index += 1
        page = page[key]
        entities_list = list(map(lambda entity: entity_class(user, entity), page))
        entities.extend(entities_list)

    return entities



import datetime

class Attempt():
    def __init__(self, attempt_id, start_time, due, step_id, status):
        self.id = attempt_id
        if isinstance(start_time, str):
            start_time = datetime.datetime.fromisoformat(start_time)
        self.start_time = start_time
        if isinstance(due, str):
            due = datetime.datetime.fromisoformat(due)
        self.due = due
        self.step_id = step_id
        self.status = status

    def __str__(self):
        attempt = "attempt: {}\tstep: {}\t".format(self.id, self.step_id)
        time_left = self.due - datetime.datetime.now()
        if time_left.seconds <= 0:
            return attempt+"status: PAST DUE"
        mins = (time_left.seconds//60)%60
        secs = time_left.seconds - 60*mins
        return attempt+"status: {}\t{} mins, {} secs left".format(self.status, mins, secs)

    def __repr__(self):
        return str(self)

    def json(self):
        attempt = vars(self).copy()
        attempt['start_time'] = str(attempt['start_time'])
        attempt['due'] = str(attempt['due'])
        return attempt



import datetime

from .settings import ATTEMPT_FILE
from .client.attempt import Attempt
from .filemanager import FileManager

_file_manager = FileManager()


def clear():
    set_data({})


def get_data():
    try:
        return _file_manager.read_json(ATTEMPT_FILE)
    except FileNotFoundError:
        clear()
        return {}


def set_data(data):
    _file_manager.write_json(ATTEMPT_FILE, data)


def get_step_id(data=None):
    if data is None:
        data = get_data()
    try:
        position = data['current_position']
        return data['steps'][position - 1]
    except KeyError:
        return None


def set_attempt(attempt, data=None):
    if data is None:
        data = get_data()
    if 'attempts' not in data:
        data['attempts'] = {}
    data['attempts'][str(attempt.step_id)] = attempt.json()
    set_data(data)


def get_attempt(step_id, data=None):
    if data is None:
        data = get_data()
    attempt = data['attempts'][str(step_id)]
    return Attempt(
        int(attempt['id']), attempt['start_time'], attempt['due'],
        int(attempt['step_id']), attempt['status']
    )


def set_lesson_id(lesson_id):
    data = get_data()
    data['lesson_id'] = lesson_id
    set_data(data)


def get_lesson_id(data=None):
    if data is None:
        data = get_data()
    try:
        return data['lesson_id']
    except KeyError:
        return None


def get_current_position(data=None):
    if data is None:
        data = get_data()
    try:
        return data['current_position']
    except KeyError:
        return None



class LanguageManager:

    programming_language = {'.cpp': 'c++11', '.c': 'c++11', '.py': 'python3',
                            '.java': 'java8', '.hs': 'haskel 7.10', '.sh': 'shell',
                            '.r': 'r', '.js': 'javascript', '.rs': 'rust', '.m': 'octave',
                            '.asm': 'asm32', '.clj': 'clojure', '.cs': 'mono c#'}

    def __init__(self):
        pass



from pathlib import Path

from .filemanager import FileManager
from .models.course import Course
from .settings import COURSE_CACHE_FILE


class CourseCache():
    def __init__(self, course=None, cache_path=COURSE_CACHE_FILE):
        self.file_manager = FileManager()
        self.path = cache_path
        self.course = course
        self._loaded = False
        if not isinstance(cache_path, Path):
            self.path = Path(self.path)
        if course is not None:
            self._initialize_empty_cache()


    def _initialize_empty_cache(self):
        self.data = {'course': [self.course.id, self.course.title], 'lessons': []}


    def load(self, user):
        """try to load the cache for a course. return success status"""
        if not self._loaded:
            try:
                data = self.file_manager.read_json(self.path)
            except FileNotFoundError:
                return False
            if self.course is None:
                # we just create a dummy Course class here, since creating the actual
                # class would require another http request
                self.course = type('Course', (), {})()
                self.course.id = data['current']
                self.course.title = ""
                self._initialize_empty_cache()
            if data['current'] != self.course.id:
                if str(self.course.id) not in data['courses']:
                    return False
                self._save_as_current(data)
            self.data['lessons'] = list(map(int, data['courses'][str(self.course.id)]['lessons']))
            self._loaded = True
        return self._loaded


    def save(self):
        try:
            data = self.file_manager.read_json(self.path)
            data['courses'][str(self.course.id)] = self.data
            data['current'] = self.course.id
        except FileNotFoundError:
            data = {'current': self.course.id, 'courses': {self.course.id: self.data}}
        self.file_manager.write_json(self.path, data)


    def _save_as_current(self, data=None):
        if data is None:
            data = self.file_manager.read_json(self.path)
        data['current'] = self.course.id
        self.file_manager.write_json(self.path, data)


    def update(self):
        """create a cache of all of the lessons in the course"""
        self.data['lessons'] = [
            lesson.id for section in self.course.items()
            for lesson in section.items()
        ]
        self.save()


    def get_next_lesson(self, lesson_id, direction, last_pos=None):
        # if the most recent position is not set, just make it be one of the extremes
        if last_pos is None:
            last_pos = (0, len(self.data['lessons'])-1)[direction < 0]
        # find the current position of this lesson in the list
        # implemented differently depending on the provided direction
        if direction > 0:
            lesson_pos = self.data['lessons'].index(lesson_id, last_pos)
        else:
            last_pos = len(self.data['lessons']) - (last_pos + 1)
            lesson_pos = self.data['lessons'][::-1].index(lesson_id, last_pos)
            lesson_pos = len(self.data['lessons']) - (lesson_pos + 1)
        # get the next lesson and its position
        lesson_pos += direction
        return self.data['lessons'][lesson_pos], lesson_pos




GRAND_TYPE_PASSWORD = "password"
GRAND_TYPE_CREDENTIALS = "client_credentials"

STEPIK_HOST = "https://stepik.org/"
STEPIK_API_URL = STEPIK_HOST + "api"

LESSONS = STEPIK_API_URL + "/lessons/"
LESSONS_PK = LESSONS + "{}"

SUBMISSIONS = STEPIK_API_URL + "/submissions/"
SUBMISSIONS_PK = SUBMISSIONS + "{}"

STEPS = STEPIK_API_URL + "/steps/"
STEPS_PK = STEPS + "{}"

COURSES = STEPIK_API_URL + "/courses/"
COURSES_PK = COURSES + "{}"

ATTEMPTS = STEPIK_API_URL + "/attempts"

SECTIONS = STEPIK_API_URL + "/sections/"
SECTIONS_PK = SECTIONS + "{}"

UNITS = STEPIK_API_URL + "/units/"



from stepik.client import stepikclient
from .entity import Entity
from .lesson import Lesson
from .unit import Unit
from ..utils import entities_loader


class Section(Entity):
    def _load(self):
        json = stepikclient.get_section(self.user, self.id)

        self._load_from_data(json['sections'][0])

    def units_set(self):
        return entities_loader(stepikclient.get_units, self.user, "units", self.units, Unit)

    def items(self):
        units = self.units_set()

        ids = list(map(lambda unit: unit.lesson, units))
        return entities_loader(stepikclient.get_lessons, self.user, "lessons", ids, Lesson)


class Entity:
    def __init__(self, user, data):
        self.user = user

        self._load_from_data(data)

    def __str__(self):
        return "{}\t{}".format(self.id, self.title)

    def _load_from_data(self, data):
        for key in data:
            self.__dict__[key] = data[key]

    def _load(self):
        pass

    @classmethod
    def get(cls, user, entity_id):
        entity = cls(user, {"id": entity_id})
        entity._load()

        return entity



from stepik.client import stepikclient
from .entity import Entity
from .section import Section
from .user import User
from ..utils import entities_loader, all_entities_loader


class Course(Entity):
    def _load(self):
        course_json = stepikclient.get_course(self.user, self.id)

        self._load_from_data(course_json['courses'][0])

    @staticmethod
    def all():
        user = User()
        return all_entities_loader(stepikclient.get_courses, user, 'courses', Section, enrolled='true')

    def items(self):
        return entities_loader(stepikclient.get_sections, self.user, "sections", self.sections, Section)




import datetime

from stepik.client.auth import auth_user_password
from stepik.client.consts import GRAND_TYPE_CREDENTIALS
from ..filemanager import FileManager
from ..settings import CLIENT_FILE, CLIENT_ID, CLIENT_SECRET


class User:
    file_manager = FileManager()

    def __init__(self):
        try:
            data = self.file_manager.read_json(CLIENT_FILE)
        except FileNotFoundError:
            data = {}

        self.client_id = data.get('client_id', CLIENT_ID)
        self.secret = data.get('client_secret', CLIENT_SECRET)
        self.username = data.get('username', 'Unknown')
        self.grand_type = data.get('grand_type', GRAND_TYPE_CREDENTIALS)
        self.access_token = data.get('access_token', '')
        self.refresh_token = data.get('refresh_token', '')
        self.expiration = data.get('expiration', datetime.datetime.min)
        if not isinstance(self.expiration, datetime.datetime):
            self.expiration = datetime.datetime.fromisoformat(self.expiration)
        self.step_type = data.get('step_type', 'all')

        if self.expiration < datetime.datetime.now() and self.client_id and self.secret:
            auth_user_password(self)

        self.save()

    def save(self):
        data = dict()
        data['client_id'] = self.client_id
        data['client_secret'] = self.secret
        data['username'] = self.username
        data['grand_type'] = self.grand_type
        data['access_token'] = self.access_token
        data['refresh_token'] = self.refresh_token
        data['expiration'] = self.expiration.isoformat()
        data['step_type'] = self.step_type

        self.file_manager.write_json(CLIENT_FILE, data)



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='stepik-cli',
    version='0.4.1',
    description="A Stepik CLI for students",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aryarm/stepik-cli",
    install_requires=[
        'click>=3.0',
        'requests',
        'colorama',
        'html2text'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(include=['stepik', 'stepik.client', 'stepik.models']),
    entry_points={
        'console_scripts': ['stepik=stepik.__main__:main']
    },
)



import os
import json

import time
import click
import requests
import datetime

from .attempt import Attempt
from .auth import get_headers
from .consts import STEPIK_API_URL, LESSONS_PK, SUBMISSIONS_PK, STEPS_PK, COURSES_PK, ATTEMPTS, SUBMISSIONS, \
    SECTIONS, UNITS, SECTIONS_PK, LESSONS, STEPS

from .. import attempt_cache
from ..filemanager import FileManager
from ..languagemanager import LanguageManager
from ..utils import exit_util, get_lesson_id, get_step_id, prepare_ids


def request(request_type, link, **kwargs):
    resp = None
    try:
        resp = requests.__dict__[request_type](link, **kwargs)
    except Exception as e:
        exit_util(e.args[0])
    if resp.status_code >= 400:
        exit_util("Something went wrong. A request returned {}".format(resp.status_code))
    return resp


def post_request(link, **kwargs):
    return request("post", link, **kwargs)


def get_request(link, **kwargs):
    return request("get", link, **kwargs)


def get_entity(user, entity_id, url_template):
    entity = get_request(url_template.format(entity_id), headers=get_headers(user))
    return entity.json()


def get_course(user, course_id):
    return get_entity(user, course_id, COURSES_PK)


def get_section(user, section_id):
    return get_entity(user, section_id, SECTIONS_PK)


def get_lesson(user, lesson_id):
    return get_entity(user, lesson_id, LESSONS_PK)


def get_submission(user, submission_id):
    return get_entity(user, submission_id, SUBMISSIONS_PK)


def get_step(user, step_id):
    return get_entity(user, step_id, STEPS_PK)


def get_attempt_id(user, step_id):
    data = json.dumps({"attempt": {"step": str(step_id)}})
    attempt = post_request(ATTEMPTS, data=data, headers=get_headers(user))
    return attempt.json()['attempts'][0]['id']


def post_submit(user, data, attempt_id):
    submission_url = SUBMISSIONS[:-1] if SUBMISSIONS.endswith('/') else SUBMISSIONS
    resp = post_request(
        submission_url+"?attempt={}".format(attempt_id),
        data=data, headers=get_headers(user)
    )
    return resp.json()


def get_languages_list(user, step_id):
    if step_id is None:
        exit_util('Please set the current step using the "step" command.')
    step = get_step(user, step_id)
    block = step['steps'][0]['block']
    if block['name'] != 'code':
        exit_util('The current step does not have a coding challenge.')
    languages = block['options']['code_templates']
    return [lang for lang in languages]


def evaluate(user, attempt_id):
    click.secho("Evaluating...", bold=True, fg='white', err=True, nl=False)
    time_out = 0.1
    max_time = 60
    while True:
        result = get_submission(user, attempt_id)
        status = result['submissions'][0]['status']
        hint = result['submissions'][0]['hint']
        if status != 'evaluation':
            break
        if time_out >= max_time:
            exit_util("\nExceeded maximum evaluation time.", 3)
        click.echo(".", nl=False, err=True)
        time.sleep(time_out)
        # time_out doubles each time (so it's exponential!)
        time_out += time_out
    click.secho("\nYour solution is {}".format(status), fg=['red', 'green'][status == 'correct'], bold=True)
    if status != 'correct':
        exit_util(hint, 2)


def get_dataset_attempt(user, step_id, attempt_id=None):
    if attempt_id is None:
        data = json.dumps({"attempt": {"step": str(step_id)}})
        try:
            attempt = post_request(ATTEMPTS, data=data, headers=get_headers(user))
            attempt = attempt.json()['attempts'][0]
        except Exception:
            exit_util(
                "Unable to attempt this challenge. Is the current position set to a valid step?\n" + \
                str(attempt)
            )
        if 'dataset_url' not in attempt or not attempt['dataset_url']:
            exit_util("Unable to retrieve dataset from current step. Check that a dataset exists for this step.")
    else:
        try:
            attempt = post_request(ATTEMPTS+"/{}".format(attempt_id), headers=get_headers(user))
            attempt = attempt.json()['attempts'][0]
        except Exception:
            exit_util(
                "Unable to attempt this challenge. Is the current position set to a valid step?\n" + \
                str(attempt)
            )
        if 'dataset_url' not in attempt or not attempt['dataset_url']:
            exit_util("Unable to retrieve dataset from this attempt. Check that a dataset exists for this attempt.")
    return attempt


def download_dataset(user, filename, step_id=None, attempt_id=None):
    data = attempt_cache.get_data()
    if step_id is None:
        step_id = attempt_cache.get_step_id(data)
    attempt = get_dataset_attempt(user, step_id, attempt_id)
    if attempt['dataset_url'].startswith('/api'):
        attempt['dataset_url'] = attempt['dataset_url'][len('/api'):]
    try:
        dataset = get_request(
            STEPIK_API_URL+attempt['dataset_url'],
            headers=get_headers(user)
        )
    except Exception:
        exit_util("Unable to retrieve dataset for attempt {}".format(attempt['id']))
    FileManager().write_to_file(filename, dataset.text)
    start_time = datetime.datetime.now()
    time_left = datetime.timedelta(seconds=int(attempt['time_left']))
    attempt = Attempt(
        attempt['id'], start_time, start_time + time_left,
        attempt['step'], attempt['status']
    )
    attempt_cache.set_attempt(attempt, data)
    return attempt


def submit_code(user, filename, lang=None, step_id=None, attempt_id=None):
    file_manager = FileManager()

    try:
        text_contents = "".join(file_manager.read_file(filename))
    except FileNotFoundError:
        exit_util("File {} not found".format(filename))

    if step_id is None:
        step_id = attempt_cache.get_step_id()
    submission = {
        "submission":
            {
                "reply":{},
                "attempt": attempt_id
            }
    }
    if lang is None:
        try:
            attempt = attempt_cache.get_attempt(step_id)
            if attempt.due < datetime.datetime.now():
                exit_util("This challenge has already expired. Create a new one with the 'dataset' command.")
        except KeyError:
            exit_util("Tried to submit to a dataset challenge, but couldn't find a prior attempt. Please specify the attempt ID via the --attempt-id parameter")
        submission['submission']['reply'] = {
            "file": text_contents
        }
        if submission['submission']['attempt'] is None:
            submission['submission']['attempt'] = attempt.id
    else:
        if attempt_id is None:
            try:
                attempt_id = get_attempt_id(user, step_id)
            except Exception:
                exit_util("Please, set the step link!")
            submission['submission']['attempt'] = attempt_id
        if lang == 'text':
            submission['submission']['reply'] = {
                "text": text_contents,
                "files": []
            }
        else:
            available_languages = get_languages_list(user, step_id)
            if lang in available_languages:
                language = lang
            else:
                language = LanguageManager().programming_language.get(os.path.splitext(filename)[1])
            if language is None:
                exit_util("Doesn't correct extension for programme.")
            if language not in available_languages:
                exit_util("This language not available for current step.")
            submission['submission']['time'] = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime())
            submission['submission']['reply'] = {
                "code": text_contents,
                "language": language
            }
    submit = post_submit(
        user, json.dumps(submission), submission['submission']['attempt']
    )
    try:
        submission = submit['submissions'][-1]['id']
    except IndexError:
        exit_util("There was a problem with the format of the submission.")
    evaluate(user, submission)


def set_step(user, step_url):
    lesson_id = get_lesson_id(step_url)
    step_id = get_step_id(step_url)

    if lesson_id is None or not step_id:
        exit_util("The link is incorrect.")

    lesson = get_lesson(user, lesson_id)

    steps = None
    try:
        steps = lesson['lessons'][0]['steps']
    except Exception:
        exit_util("Didn't receive such lesson.")
    if len(steps) < step_id or step_id < 1:
        exit_util("Too few steps in the lesson.")

    data = attempt_cache.get_data()
    data['steps'] = steps
    data['current_position'] = step_id
    attempt_cache.set_data(data)
    try:
        attempt_cache.set_lesson_id(lesson_id)
    except PermissionError:
        exit_util("You do not have permission to perform this action.")
    click.secho("Current position set to lesson {}, step {}".format(lesson_id, step_id), fg="green", bold=True)


def get_courses(user, **kwargs):
    courses = get_request(STEPIK_API_URL + "/courses/", params=kwargs, headers=get_headers(user))
    return courses.json()


def get_entities_with_ids(user, ids, page, url):
    url = url + "?" + prepare_ids(ids) + '&page=' + str(page)
    entities = get_request(url, headers=get_headers(user))
    return entities.json()


def get_sections(user, ids, page):
    return get_entities_with_ids(user, ids, page, SECTIONS)


def get_units(user, ids, page):
    return get_entities_with_ids(user, ids, page, UNITS)


def get_lessons(user, ids, page):
    return get_entities_with_ids(user, ids, page, LESSONS)


def get_steps(user, ids, page):
    return get_entities_with_ids(user, ids, page, STEPS)



from stepik.client import stepikclient
from .entity import Entity
from .step import Step
from ..utils import entities_loader


class Lesson(Entity):
    def _load(self):
        json = stepikclient.get_lesson(self.user, self.id)

        self._load_from_data(json['lessons'][0])

    def items(self):
        return entities_loader(stepikclient.get_steps, self.user, "steps", self.steps, Step)



from . import attempt_cache
from .utils import exit_util
from .models.lesson import Lesson
from .models.section import Section
from .course_cache import CourseCache

FORWARD = 1
BACK = -1

cached_lessons = CourseCache()

def update_step_position(position, steps, direction):
    position += direction
    if position > len(steps) or position < 1:
        # if we get here, it means we've run out of steps in this lesson!
        raise ValueError
    return position

def navigate(user, step_type, direction, course_cache=None):
    data = attempt_cache.get_data()
    try:
        position = data['current_position']
        lesson = Lesson.get(user, data['lesson_id'])
        steps = lesson.items()
    except KeyError:
        return False

    # ensure that direction is either -1 or 1
    # idk why we have to do this, but it was in the original code
    direction = (-1,1)[direction > 0]

    lesson_pos = None
    while True:
        try:
            position = update_step_position(position, steps, direction)
        except ValueError:
            if course_cache is None:
                break
            else:
                try:
                    lesson_id, lesson_pos = course_cache.get_next_lesson(
                        lesson.id, direction, lesson_pos
                    )
                except ValueError:
                    break
                lesson = Lesson.get(user, lesson_id)
                data['lesson_id'] = lesson.id
                steps = lesson.items()
                data['steps'] = [s.id for s in steps]
                position = (0, len(steps)+1)[direction < 0]
                continue
        step = steps[position-1]
        if step_type == "all" or step.block['name'] == step_type:
            data['current_position'] = position
            attempt_cache.set_data(data)
            return True
    return False


def next_step(user, step_type):
    if not cached_lessons.load(user):
        exit_util("Please first set the course ID via the 'course' command.")
    return navigate(user, step_type, FORWARD, cached_lessons)


def prev_step(user, step_type):
    if not cached_lessons.load(user):
        exit_util("Please first set the course ID via the 'course' command.")
    return navigate(user, step_type, BACK, cached_lessons)

def create_course_cache(course):
    global cached_lessons
    cached_lessons = CourseCache(course)
    return cached_lessons



import os
import sys
import json
from pathlib import Path


class FileManager:
    """
    Local file manager
    """

    def __init__(self):
        self.home = os.getcwd()

    def create_dir(self, dir_name):
        dir_name = self.get_name(dir_name)
        try:
            os.mkdir(dir_name)
        except FileExistsError as e:
            return

    def get_name(self, filename):
        if isinstance(filename, Path):
            return str(filename)
        return filename

    def read_file(self, filename):
        filename = self.get_name(filename)
        with (
            open(filename, "r") if filename and filename != '-' else sys.stdin
        ) as file:
            for line in file:
                yield line

    def write_to_file(self, filename, content):
        filename = self.get_name(filename)
        with (
            open(filename, "w") if filename and filename != '-' else sys.stdout
        ) as file:
            file.writelines(content)

    def write_json(self, filename, data):
        filename = self.get_name(filename)
        with (
            open(filename, "w") if filename and filename != '-' else sys.stdout
        ) as file:
            json.dump(data, file)

    def read_json(self, filename):
        filename = self.get_name(filename)
        return json.loads((
            open(filename) if filename and filename != '-' else sys.stdin
        ).read())

    @staticmethod
    def is_local_file(filename):
        return os.path.isfile(filename)



import os 

APP_FOLDER = os.path.dirname(os.path.realpath(__file__))+"/.stepik"
COURSE_CACHE_FILE = APP_FOLDER + "/course_cache_file"
CLIENT_FILE = APP_FOLDER + "/client_file"
ATTEMPT_FILE = APP_FOLDER + "/attempt_file"
CLIENT_ID = ""
CLIENT_SECRET = ""
